--==============================================================================
-- CERN (BE-CO-HT)
-- Converter board common gateware top-level file
--==============================================================================
--
-- author: Theodor Stana (t.stana@cern.ch)
--
-- date of creation: 2014-07-24
--
-- version: 1.0
--
-- description:
--    This module is to be instantiated in all pulse converter board designs,
--    for example the CONV-TTL-BLO or CONV-TTL-RS485. It contains a set of
--    modules common to all these boards, that are configurable via generics to
--    accommodate for the application of each board.
--
-- dependencies:
--    general-cores repository [1]
--
-- references:
--   [1] Board IDs for level conversion circuits
--       www.ohwr.org/projects/conv-common-gw/wiki/Board-id
--   [2] Platform-independent core collection webpage on OHWR,
--       http://www.ohwr.org/projects/general-cores/repository
--   [3] ELMA, Access to board data using SNMP and I2C
--       http://www.ohwr.org/documents/227
--
--==============================================================================
-- GNU LESSER GENERAL PUBLIC LICENSE
--==============================================================================
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--==============================================================================
-- last changes:
--    2014-07-24   Theodor Stana     File created
--==============================================================================
-- TODO: -
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

use work.gencores_pkg.all;
use work.wishbone_pkg.all;
use work.genram_pkg.all;
use work.wrcore_pkg.all;
use work.wr_xilinx_pkg.all;

use work.conv_common_gw_pkg.all;

entity conv_common_gw is
  generic
  (
    -- Number of repeater channels
    g_nr_chans             : integer := 6;

    -- Board ID -- 4-letter ASCII string indicating the board ID
    -- see [1] for example
    g_board_id             : std_logic_vector(31 downto 0);
    -- Gateware version
    g_gwvers               : std_logic_vector(7 downto 0);

    -- Generate pulse repetition logic with fixed output pulse width
    g_pgen_fixed_width     : boolean;
    -- Pulse width at pulse generator output (valid with fixed output pulse width)
    g_pgen_pwidth          : natural := 24;
    -- Duty cycle divider ratio for pulse generator
    -- output pulse will be limited to 1/g_pgen_duty_cycle_div
    g_pgen_duty_cycle_div  : natural := 5;
    -- Pulse generator glitch filter length in number of clk_20_i cycles
    g_pgen_gf_len          : integer := 4;

    -- Generate logic with pulse counters
    g_with_pulse_cnt       : boolean := false;

    -- Generate logic with pulse counters
    g_with_pulse_timetag   : boolean := false;

    -- Generate logic with manual trigger
    g_with_man_trig        : boolean := false;
    g_man_trig_pwidth      : integer := 24;

    -- Generate one-wire master for thermometer
    g_with_thermometer     : boolean := false;

    -- Generate logic for White Rabbit implementation
    g_with_wr              : boolean := false;

    -- Bicolor LED controller signals
    g_bicolor_led_columns  : integer := 6;
    g_bicolor_led_lines    : integer := 2
  );
  port
  (
    -- Clocks
    clk_20_i                : in  std_logic;
    clk_125_p_i             : in  std_logic;
    clk_125_n_i             : in  std_logic;

    -- Reset output signal, synchronous to 20 MHz clock
    rst_n_o                 : out std_logic;

    -- Glitch filter active-low enable signal
    gf_en_n_i               : in  std_logic;

    -- Channel enable
    global_ch_oen_o         : out std_logic;
    pulse_front_oen_o       : out std_logic;
    pulse_rear_oen_o        : out std_logic;
    inv_oen_o               : out std_logic;

    -- Pulse inputs
    pulse_i                 : in  std_logic_vector(g_nr_chans-1 downto 0);
    pulse_o                 : out std_logic_vector(g_nr_chans-1 downto 0);

    -- Channel leds
    -- 26 ms active-high pulse on pulse_o rising edge
    led_pulse_o             : out std_logic_vector(g_nr_chans-1 downto 0);

    -- I2C interface
    scl_i                   : in  std_logic;
    scl_o                   : out std_logic;
    scl_en_o                : out std_logic;
    sda_i                   : in  std_logic;
    sda_o                   : out std_logic;
    sda_en_o                : out std_logic;

    -- I2C LED signals -- conect to a bicolor LED of choice
    -- led_i2c_o pulses four times on I2C transfer
    led_i2c_o               : out std_logic;

    -- VME interface
    vme_sysreset_n_i        : in  std_logic;
    vme_ga_i                : in  std_logic_vector(4 downto 0);
    vme_gap_i               : in  std_logic;

    -- SPI interface to on-board flash chip
    flash_cs_n_o            : out std_logic;
    flash_sclk_o            : out std_logic;
    flash_mosi_o            : out std_logic;
    flash_miso_i            : in  std_logic;

    -- PLL DACs
    -- 20 MHz VCXO control
    dac20_din_o             : out std_logic;
    dac20_sclk_o            : out std_logic;
    dac20_sync_n_o          : out std_logic;
    -- 125 MHz clock generator control
    dac125_din_o            : out std_logic;
    dac125_sclk_o           : out std_logic;
    dac125_sync_n_o         : out std_logic;

    -- SFP lines
    sfp_los_i               : in    std_logic;
    sfp_present_i           : in    std_logic;
    sfp_rate_select_o       : out   std_logic;
    sfp_scl_b               : inout std_logic;
    sfp_sda_b               : inout std_logic;
    sfp_tx_disable_o        : out   std_logic;
    sfp_tx_fault_i          : in    std_logic;

    -- White rabbit LED lines
    wr_led_link_o           : out   std_logic;
    wr_led_act_o            : out   std_logic;

    -- FPGA Multi-Gigabit Transceiver connections
    mgt_refclk0_p_i         : in    std_logic;
    mgt_refclk0_n_i         : in    std_logic;
    mgt_tx0_p_o             : out   std_logic;
    mgt_tx0_n_o             : out   std_logic;
    mgt_rx0_p_i             : in    std_logic;
    mgt_rx0_n_i             : in    std_logic;

    -- Switch inputs (for readout from converter status register)
    sw_gp_i                 : in  std_logic_vector(7 downto 0);
    sw_other_i              : in  std_logic_vector(31 downto 0);

    -- RTM lines
    rtmm_i                  : in  std_logic_vector(2 downto 0);
    rtmp_i                  : in  std_logic_vector(2 downto 0);

    -- TTL, INV-TTL and rear-panel channel inputs, for reflection in line status register
    line_front_i            : in std_logic_vector(g_nr_chans-1 downto 0);
    line_inv_i              : in std_logic_vector(3 downto 0);
    line_rear_i             : in std_logic_vector(g_nr_chans-1 downto 0);

    -- Fail-safe lines, detect invalid or no signal on channel input
    line_front_fs_i         : in std_logic_vector(g_nr_chans-1 downto 0);
    line_inv_fs_i           : in std_logic_vector(3 downto 0);
    line_rear_fs_i          : in std_logic_vector(g_nr_chans-1 downto 0);

    -- Thermometer line
    thermometer_b           : inout std_logic;

    -- System error LED, active-high on system error
    -- ERR bicolor LED should light red when led_syserr_o = '1'
    led_syserr_o            : out std_logic;

    -- Bicolor LED signals
    bicolor_led_state_i     : in  std_logic_vector(2*g_bicolor_led_columns*g_bicolor_led_lines-1 downto 0);
    bicolor_led_col_o       : out std_logic_vector(g_bicolor_led_columns-1 downto 0);
    bicolor_led_line_o      : out std_logic_vector(g_bicolor_led_lines-1 downto 0);
    bicolor_led_line_oen_o  : out std_logic_vector(g_bicolor_led_lines-1 downto 0)
  );
end entity conv_common_gw;


architecture arch of conv_common_gw is

  --============================================================================
  -- Constant declarations
  --============================================================================
  -- Number of Wishbone masters and slaves, for wb_crossbar
  constant c_nr_masters       : natural :=  1;
  constant c_nr_slaves        : natural :=  2;

  -- slave order definitions
  constant c_slv_conv_regs    : natural := 0;
  constant c_slv_multiboot    : natural := 1;

  -- base address definitions
  constant c_addr_conv_regs   : t_wishbone_address := x"00000000";
  constant c_addr_multiboot   : t_wishbone_address := x"00000100";
  constant c_addr_sdb         : t_wishbone_address := x"00000f00";

  -- SDB interconnect layout
  -- c_conv_regs_sdb defined in conv_common_gw_pkg.vhd
  constant c_sdb_layout : t_sdb_record_array(c_nr_slaves-1 downto 0) := (
      c_slv_conv_regs => f_sdb_embed_device(c_conv_regs_sdb, c_addr_conv_regs),
      c_slv_multiboot => f_sdb_embed_device(c_xwb_xil_multiboot_sdb,
                              c_addr_multiboot)
    );

  -- Tag bufferdata width:  40 -- TAI
  --                        28 -- cycles
  --                         1 -- WRPRES bit
  --                        xx -- channel mask for max. nr. channels
  constant c_tagbuff_data_width    : positive := 40 + 28 + 1 + c_max_nr_chans;

  --============================================================================
  -- Type declarations
  --============================================================================
  -- Max. channel count of c_max_nr_chans enforced here:
  type t_pulse_led_cnt is array(c_max_nr_chans-1 downto 0)
                            of unsigned(18 downto 0);
  type t_pulse_cnt is array(c_max_nr_chans-1 downto 0)
                        of unsigned(31 downto 0);
  type t_ch_pcr is array(c_max_nr_chans-1 downto 0)
                    of std_logic_vector(31 downto 0);

  -- Latest timestamp
  type t_lts_tai is array(c_max_nr_chans-1 downto 0)
                                  of std_logic_vector(39 downto 0);
  type t_lts_cycles is array(c_max_nr_chans-1 downto 0)
                                      of std_logic_vector(27 downto 0);

  --============================================================================
  -- Component declarations
  --============================================================================
  component serialIdTempInt is
    generic (
      FREQ : integer := 40 --Frequency in MHz
      );
    port(
      Clk  : in std_logic;
      RstN : in std_logic;

      SerialId : inout std_logic;
      Id       : out   std_logic_vector(63 downto 0);
      Temp     : out   std_logic_vector(15 downto 0);
      IdRead   : out   std_logic;
      Pps      : in    std_logic;
      IdOk     : out   std_logic
      );
  end component serialIdTempInt;

  --============================================================================
  -- Signal declarations
  --============================================================================
  -- Per-domain clock and reset signals
  signal clk_125             : std_logic;
  signal clk_dmtd            : std_logic;
  signal clk_dmtd_fr_pll     : std_logic;
  signal clk_dmtd_pll_fb     : std_logic;
  signal clk_20_buf          : std_logic;
  signal clk_gtp             : std_logic;
  signal rst_125_n           : std_logic;
  signal rst_20_n            : std_logic;
  signal rst_ext             : std_logic;

  -- Pulse logic signals
  signal trig_a              : std_logic_vector(g_nr_chans-1 downto 0);
  signal trig_synced         : std_logic_vector(g_nr_chans-1 downto 0);
  signal trig_degl           : std_logic_vector(g_nr_chans-1 downto 0);
  signal trig_chan           : std_logic_vector(g_nr_chans-1 downto 0);
  signal trig_chan_redge_p   : std_logic_vector(g_nr_chans-1 downto 0);
  signal trig_man            : std_logic_vector(g_nr_chans-1 downto 0);
  signal trig_pgen           : std_logic_vector(g_nr_chans-1 downto 0);
  signal pulse_outp          : std_logic_vector(g_nr_chans-1 downto 0);
  signal pulse_outp_d0       : std_logic_vector(g_nr_chans-1 downto 0);
  signal pulse_outp_redge_p  : std_logic_vector(g_nr_chans-1 downto 0);
  signal pmisse_p            : std_logic_vector(g_nr_chans-1 downto 0);

  -- Output enable signals
  signal global_oen          : std_logic;
  signal ttl_oen, invttl_oen : std_logic;
  signal rear_oen            : std_logic;

  -- I2C bridge signals
  signal i2c_addr            : std_logic_vector(6 downto 0);
  signal i2c_tip             : std_logic;
  signal i2c_err_p           : std_logic;
  signal i2c_wdto_p          : std_logic;

  -- Signals to/from converter system registers component
  signal rtm_lines           : std_logic_vector(5 downto 0);
  signal sw_gp               : std_logic_vector(7 downto 0);
  signal sw_multicast        : std_logic_vector(3 downto 0);
  signal rst_unlock          : std_logic;
  signal rst_unlock_bit      : std_logic;
  signal rst_unlock_bit_ld   : std_logic;
  signal rst_bit             : std_logic;
  signal rst_bit_ld          : std_logic;
  signal rst_fr_reg          : std_logic;
  signal i2c_wdto_bit        : std_logic;
  signal i2c_wdto_bit_rst    : std_logic;
  signal i2c_wdto_bit_rst_ld : std_logic;
  signal pmisse_bit          : std_logic_vector(c_max_nr_chans-1 downto 0);
  signal pmisse_bit_rst      : std_logic_vector(c_max_nr_chans-1 downto 0);
  signal pmisse_bit_rst_ld   : std_logic;
  signal pmisse_bits_or      : std_logic;
  signal pulse_cnt           : t_pulse_cnt;
  signal ch_pcr              : t_ch_pcr;
  signal ch_pcr_ld           : std_logic_vector(c_max_nr_chans-1 downto 0);
  signal mpt_ld              : std_logic;
  signal mpt                 : std_logic_vector( 7 downto 0);
  signal tvlr                : std_logic_vector(31 downto 0);
  signal tvlr_ld             : std_logic;
  signal tvhr                : std_logic_vector( 7 downto 0);
  signal tvhr_ld             : std_logic;
  signal wrpres              : std_logic;
  signal i2c_err_bit         : std_logic;
  signal i2c_err_bit_rst     : std_logic;
  signal i2c_err_bit_rst_ld  : std_logic;
  signal line_front          : std_logic_vector(c_max_nr_chans-1 downto 0);
  signal line_rear           : std_logic_vector(c_max_nr_chans-1 downto 0);
  signal line_front_fs       : std_logic_vector(c_max_nr_chans-1 downto 0);
  signal line_rear_fs        : std_logic_vector(c_max_nr_chans-1 downto 0);

  -- LED signals
  signal led_pulse           : std_logic_vector(g_nr_chans-1 downto 0);
  signal led_pulse_cnt       : t_pulse_led_cnt;
  signal led_i2c             : std_logic;
  signal led_i2c_clkdiv      : unsigned(18 downto 0);
  signal led_i2c_cnt         : unsigned( 2 downto 0);
  signal led_i2c_blink       : std_logic;

  -- Wishbone crossbar signals
  signal xbar_slave_in       : t_wishbone_slave_in_array  (c_nr_masters-1 downto 0);
  signal xbar_slave_out      : t_wishbone_slave_out_array (c_nr_masters-1 downto 0);
  signal xbar_master_in      : t_wishbone_master_in_array (c_nr_slaves-1 downto 0);
  signal xbar_master_out     : t_wishbone_master_out_array(c_nr_slaves-1 downto 0);

  -- Time-tagging component signals
  signal tm_cycles            : std_logic_vector(27 downto 0);
  signal tm_tai               : std_logic_vector(39 downto 0);

  signal buf_wr_req_p         : std_logic;
  signal buf_rd_req_p         : std_logic;
  signal buf_count            : std_logic_vector(f_log2_size(128)-1 downto 0);
  signal buf_full             : std_logic;
  signal buf_empty            : std_logic;
  signal buf_chan             : std_logic_vector(g_nr_chans-1 downto 0);
  signal buf_wrtag            : std_logic;
  signal buf_clr_bit          : std_logic;
  signal buf_clr_bit_ld       : std_logic;
  signal buf_clr_p            : std_logic;
  signal buf_dat_in           : std_logic_vector(c_tagbuff_data_width-1 downto 0);
  signal buf_dat_out          : std_logic_vector(c_tagbuff_data_width-1 downto 0);

  -- Latest timestamp signals
  signal lts_ld_125           : std_logic_vector(c_max_nr_chans-1 downto 0);
  signal lts_ld_rdy_125       : std_logic_vector(c_max_nr_chans-1 downto 0);
  signal lts_tai_125          : t_lts_tai;
  signal lts_cycles_125       : t_lts_cycles;
  signal lts_wrtag_125        : std_logic_vector(c_max_nr_chans-1 downto 0);
  signal lts_ld_toggle        : std_logic_vector(c_max_nr_chans-1 downto 0);
  signal lts_ld_toggle_d0     : std_logic_vector(c_max_nr_chans-1 downto 0);
  signal lts_ld_toggle_d1     : std_logic_vector(c_max_nr_chans-1 downto 0);
  signal lts_ld_20            : std_logic_vector(c_max_nr_chans-1 downto 0);
  signal lts_tai_20           : t_lts_tai;
  signal lts_cycles_20        : t_lts_cycles;
  signal lts_wrtag_20         : std_logic_vector(c_max_nr_chans-1 downto 0);

  -- Temperature sensor signals
  signal tempr                : std_logic_vector(15 downto 0);
  signal uidr                 : std_logic_vector(63 downto 0);
  signal pps_p                : std_logic;
  signal pps_cnt              : unsigned(24 downto 0);

  -- MGT phy signals
  signal phy_tx_data          : std_logic_vector(7 downto 0);
  signal phy_tx_k             : std_logic;
  signal phy_tx_disparity     : std_logic;
  signal phy_tx_enc_err       : std_logic;
  signal phy_rx_data          : std_logic_vector(7 downto 0);
  signal phy_rx_rbclk         : std_logic;
  signal phy_rx_k             : std_logic;
  signal phy_rx_enc_err       : std_logic;
  signal phy_rx_bitslide      : std_logic_vector(3 downto 0);
  signal phy_rst              : std_logic;
  signal phy_loopen           : std_logic;

  -- SFP I2C signals
  signal sfp_scl_fr_wrpc      : std_logic;
  signal sfp_scl_to_wrpc      : std_logic;
  signal sfp_sda_fr_wrpc      : std_logic;
  signal sfp_sda_to_wrpc      : std_logic;

  -- DAC signals
  signal dac_dpll_load_p      : std_logic;
  signal dac_dpll_data        : std_logic_vector(15 downto 0);
  signal dac_hpll_load_p      : std_logic;
  signal dac_hpll_data        : std_logic_vector(15 downto 0);

--==============================================================================
--  architecture begin
--==============================================================================
begin

  --============================================================================
  -- Differential input buffer for 125 MHz clock
  --============================================================================
  cmp_clk_125_buf : IBUFGDS
    generic map (
      DIFF_TERM    => true,  -- Differential Termination
      IBUF_LOW_PWR => true,  -- Low power (TRUE) vs. performance (FALSE) setting
                             -- for referenced I/O standards
      IOSTANDARD   => "DEFAULT")
    port map (
      O  => clk_125,
      I  => clk_125_p_i,
      IB => clk_125_n_i
      );

  --============================================================================
  -- Internal and external reset generation
  --============================================================================
  -- External reset input to reset generator
  rst_ext <= rst_fr_reg or (not vme_sysreset_n_i);

  -- Configure reset generator for 100ms reset
  cmp_reset_gen : conv_reset_gen
    generic map
    (
      -- Reset time: 50ns * 2 * (10**6) = 100 ms
      g_reset_time => 2*(10**6)
    )
    port map
    (
      clk_i   => clk_20_i,
      rst_i   => rst_ext,
      rst_n_o => rst_20_n
    );

  -- Output reset signal is synchronous to the 20 MHz clock
  rst_n_o <= rst_20_n;

  -- And synchronize the 20 MHz domain reset into the 125 MHz domain
  cmp_sync_rst : gc_sync_ffs
   generic map
   (
     g_sync_edge => "positive"
   )
   port map
   (
     clk_i    => clk_125,
     rst_n_i  => '1',
     data_i   => rst_20_n,
     synced_o => rst_125_n
   );

  --============================================================================
  -- Output enable logic
  --============================================================================
  -- Enable outputs only when the FPGA is ready to handle them
  -- One clock cycle delay from global OEN to rest of OENs
  p_delay_oen : process (clk_20_i) is
  begin
    if rising_edge(clk_20_i) then
      if (rst_20_n = '0') then
        global_oen <= '0';
        ttl_oen    <= '0';
        invttl_oen <= '0';
        rear_oen   <= '0';
      else
        global_oen <= '1';
        if global_oen = '1' then
          ttl_oen    <= '1';
          invttl_oen <= '1';
          rear_oen   <= '1';
        end if;
      end if;
    end if;
  end process p_delay_oen;

  -- Assign OEN outputs
  global_ch_oen_o   <= global_oen;
  pulse_front_oen_o <= ttl_oen;
  inv_oen_o         <= invttl_oen;
  pulse_rear_oen_o  <= rear_oen;

  --============================================================================
  -- Pulse repetition logic
  --============================================================================
  trig_a <= pulse_i;

--------------------------------------------------------------------------------
gen_man_trig : if (g_with_man_trig = true) generate
  -- Manual trigger logic
  cmp_man_trig : conv_man_trig
    generic map
    (
      g_nr_chan => g_nr_chans,
      g_pwidth  => g_man_trig_pwidth
    )
    port map
    (
      clk_i    => clk_20_i,
      rst_n_i  => rst_20_n,
      reg_ld_i => mpt_ld,
      reg_i    => mpt,
      trig_o   => trig_man
    );
end generate gen_man_trig;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
gen_no_man_trig : if (g_with_man_trig = false) generate
  trig_man <= (others => '0');
end generate gen_no_man_trig;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
gen_pulse_chan_logic : for i in 0 to g_nr_chans-1 generate

  -- Synchronize the asynchronous trigger input into the 20 MHz clock
  -- domain before passing it to the glitch filter
  cmp_trig_sync : gc_sync_ffs
    generic map
    (
      g_sync_edge => "positive"
    )
    port map
    (
      clk_i    => clk_20_i,
      rst_n_i  => rst_20_n,
      data_i   => trig_a(i),
      synced_o => trig_synced(i)
    );

  -- Deglitch synchronized trigger signal
  cmp_inp_glitch_filt : gc_glitch_filt
    generic map
    (
      g_len => g_pgen_gf_len
    )
    port map
    (
      clk_i   => clk_20_i,
      rst_n_i => rst_20_n,
      dat_i   => trig_synced(i),
      dat_o   => trig_degl(i)
    );

  -- Now that we have a deglitched signal, generate the MUX to select between
  -- deglitched and direct channel input
  trig_chan(i) <= trig_a(i) when (gf_en_n_i = '1') else
                  trig_degl(i);

  -- Now, sync this channel trigger signal before passing it to the counters
  --
  -- The pulse counter is triggered only by a pulse that actually makes it
  -- to the pulse generator.
  --
  -- NOTE: glitch-filtered signal is also synced in 20MHz clock domain, but
  -- another sync chain here avoids extra logic complication and shoudl have
  -- no influence on the correctness of the pulse counter value
  cmp_sync_ffs : gc_sync_ffs
    port map
    (
      clk_i    => clk_20_i,
      rst_n_i  => rst_20_n,
      data_i   => trig_chan(i),
      ppulse_o => trig_chan_redge_p(i)
    );

--------------------------------------------------------------------------------
gen_pulse_cnt : if (g_with_pulse_cnt = true) generate

  -- First, the pulse counters for the used channes (up to g_nr_chans)
  p_pulse_cnt : process (clk_20_i)
  begin
    if rising_edge(clk_20_i) then
      if (rst_20_n = '0') then
        pulse_cnt(i) <= (others => '0');
      elsif (ch_pcr_ld(i) = '1') then
        pulse_cnt(i) <= unsigned(ch_pcr(i));
      elsif (trig_chan_redge_p(i) = '1') then
        pulse_cnt(i) <= pulse_cnt(i) + 1;
      end if;
    end if;
  end process p_pulse_cnt;

--------------------------------------------------------------------------------
-- Connect pulse counter values for unused channels to all zeroes
gen_pulse_cnt_unused_chans : if (g_nr_chans < c_max_nr_chans) generate
  pulse_cnt(c_max_nr_chans-1 downto g_nr_chans) <= (others => (others => '0'));
end generate gen_pulse_cnt_unused_chans;
--------------------------------------------------------------------------------

end generate gen_pulse_cnt;
--------------------------------------------------------------------------------

  -- The trigger to the pulse generator is either manual OR from the channel input
  trig_pgen(i) <= trig_chan(i) or trig_man(i);

  -- Instantiate pulse generator block for the channel
  cmp_pulse_gen : conv_pulse_gen
    generic map
    (
      g_with_fixed_pwidth => g_pgen_fixed_width,
      g_pwidth            => g_pgen_pwidth,
      g_duty_cycle_div    => g_pgen_duty_cycle_div
    )
    port map
    (
      clk_i         => clk_20_i,
      rst_n_i       => rst_20_n,

      gf_en_n_i     => gf_en_n_i,

      en_i          => '1',

      trig_a_i      => trig_pgen(i),

      pulse_err_p_o => pmisse_p(i),

      pulse_o       => pulse_outp(i)
    );

  -- Process to flash pulse LED when a pulse is output
  -- LED flash length: 26 ms
  p_pulse_led : process (clk_20_i) is
  begin
    if rising_edge(clk_20_i) then
      if (rst_20_n = '0') then
        pulse_outp_d0(i)      <= '0';
        pulse_outp_redge_p(i) <= '0';
        led_pulse_cnt(i)      <= (others => '0');
        led_pulse(i)          <= '0';
      else
        pulse_outp_d0(i)      <= pulse_outp(i);
        pulse_outp_redge_p(i) <= pulse_outp(i) and (not pulse_outp_d0(i));
        case led_pulse(i) is
          when '0' =>
            if (pulse_outp_redge_p(i) = '1') then
              led_pulse(i) <= '1';
            end if;
          when '1' =>
            led_pulse_cnt(i) <= led_pulse_cnt(i) + 1;
            if (led_pulse_cnt(i) = (led_pulse_cnt(i)'range => '1')) then
              led_pulse(i) <= '0';
            end if;
          when others =>
            led_pulse(i) <= '0';
        end case;
      end if;
    end if;
  end process p_pulse_led;

end generate gen_pulse_chan_logic;
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
gen_pulse_timetag : if (g_with_pulse_timetag = true) generate
  cmp_pulse_timetag : conv_pulse_timetag
    generic map
    (
    -- Frequency in Hz of the clk_i signal
      g_clk_rate => 125000000,
    -- Number of repetition channels
      g_nr_chan  => g_nr_chans
    )
    port map
    (
    -- Clock and active-low reset
      clk_i            => clk_125,
      rst_n_i          => rst_125_n,

    -- Asynchronous pulse input
      pulse_a_i        => trig_chan,

    -- Time inputs from White Rabbit
      wr_tm_cycles_i   => (others => '0'),
      wr_tm_tai_i      => (others => '0'),
      wr_tm_valid_i    => '0',

    -- Timing inputs from Wishbone-mapped registers
      wb_tm_tai_l_i    => tvlr,
      wb_tm_tai_l_ld_i => tvlr_ld,
      wb_tm_tai_h_i    => tvhr,
      wb_tm_tai_h_ld_i => tvhr_ld,

    -- Timing outputs
      tm_cycles_o      => tm_cycles,
      tm_tai_o         => tm_tai,
      tm_wrpres_o      => buf_wrtag,
      chan_p_o         => buf_chan,

    -- Ring buffer I/O
      buf_wr_req_p_o   => buf_wr_req_p
    );

--------------------------------------------------------------------------------
gen_buf_chan : if (g_nr_chans = c_max_nr_chans) generate
  buf_dat_in(c_max_nr_chans-1 downto  0) <= buf_chan;
end generate gen_buf_chan;

gen_buf_chan_unused_chans: if (g_nr_chans < c_max_nr_chans) generate
  buf_dat_in(g_nr_chans-1 downto 0) <= buf_chan;
  buf_dat_in(c_max_nr_chans-1 downto g_nr_chans) <= (others => '0');
end generate gen_buf_chan_unused_chans;
--------------------------------------------------------------------------------

  buf_dat_in( 6)           <= buf_wrtag;
  buf_dat_in(34 downto  7) <= tm_cycles;
  buf_dat_in(74 downto 35) <= tm_tai;

  -- Instantiate the ring buffer
  cmp_ring_buf : conv_ring_buf
    generic map
    (
      g_data_width => c_tagbuff_data_width,
      g_size       => 128
    )
    port map
    (
      -- Clocks and reset
      clk_rd_i       => clk_20_i,
      clk_wr_i       => clk_125,
      rst_n_a_i      => rst_20_n,

      -- Buffer inputs
      buf_dat_i      => buf_dat_in,
      buf_rd_req_i   => buf_rd_req_p,
      buf_wr_req_i   => buf_wr_req_p,
      buf_clr_i      => buf_clr_p,

      -- Buffer outputs
      buf_dat_o      => buf_dat_out,
      buf_full_o     => buf_full,
      buf_empty_o    => buf_empty,
      buf_count_o    => buf_count
    );

end generate gen_pulse_timetag;
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

  -- Channel output assignments
  pulse_o     <= pulse_outp;
  led_pulse_o <= led_pulse;

  --============================================================================
  -- I2C bridge logic
  --============================================================================
  -- Set the I2C address signal according to ELMA protocol [2]
  i2c_addr <= "10" & vme_ga_i;

  -- Instantiate I2C bridge component
  --
  -- FSM watchdog timeout timer:
  -- * consider bit period of 30 us
  -- * 10 bits / byte transfer => 300 us
  -- * 40 bytes in one transfer => 12000 us
  -- * clk_i period = 50 ns => g_fsm_wdt = 12000 us / 50 ns = 240000
  -- * multiply by two for extra safety => g_fsm_wdt = 480000
  -- * Time to watchdog timeout: 480000 * 50ns = 24 ms
  cmp_i2c_bridge : wb_i2c_bridge
    generic map
    (
      g_fsm_wdt => 480000
    )
    port map
    (
      -- Clock, reset
      clk_i      => clk_20_i,
      rst_n_i    => rst_20_n,

      -- I2C lines
      scl_i      => scl_i,
      scl_o      => scl_o,
      scl_en_o   => scl_en_o,
      sda_i      => sda_i,
      sda_o      => sda_o,
      sda_en_o   => sda_en_o,

      -- I2C address and status
      i2c_addr_i => i2c_addr,

      -- TIP and ERR outputs
      tip_o      => i2c_tip,
      err_p_o    => i2c_err_p,
      wdto_p_o   => i2c_wdto_p,

      -- Wishbone master signals
      wbm_stb_o  => xbar_slave_in(0).stb,
      wbm_cyc_o  => xbar_slave_in(0).cyc,
      wbm_sel_o  => xbar_slave_in(0).sel,
      wbm_we_o   => xbar_slave_in(0).we,
      wbm_dat_i  => xbar_slave_out(0).dat,
      wbm_dat_o  => xbar_slave_in(0).dat,
      wbm_adr_o  => xbar_slave_in(0).adr,
      wbm_ack_i  => xbar_slave_out(0).ack,
      wbm_rty_i  => xbar_slave_out(0).rty,
      wbm_err_i  => xbar_slave_out(0).err
    );

  -- Process to blink the LED when an I2C transfer is in progress
  -- blinks four times per transfer
  -- blink width : 20 ms
  -- blink period: 40 ms
  p_i2c_blink : process(clk_20_i)
  begin
    if rising_edge(clk_20_i) then
      if (rst_20_n = '0') then
        led_i2c_clkdiv <= (others => '0');
        led_i2c_cnt    <= (others => '0');
        led_i2c        <= '0';
        led_i2c_blink  <= '0';
      else
        case led_i2c_blink is

          when '0' =>
            led_i2c <= '0';
            if (i2c_tip = '1') then
              led_i2c_blink <= '1';
            end if;

          when '1' =>
            led_i2c_clkdiv <= led_i2c_clkdiv + 1;
            if (led_i2c_clkdiv = 399999) then
              led_i2c_clkdiv <= (others => '0');
              led_i2c_cnt    <= led_i2c_cnt + 1;
              led_i2c        <= not led_i2c;
              if (led_i2c_cnt = 7) then
                led_i2c_cnt <= (others => '0');
                led_i2c_blink <= '0';
              end if;
            end if;

          when others =>
            led_i2c_blink <= '0';

        end case;
      end if;
    end if;
  end process p_i2c_blink;

  -- And set the I2C LED Output
  led_i2c_o <= led_i2c;

  -- Register for the I2C_WDTO bit in the SR, cleared by writing a '1'
  p_sr_wdto_bit : process(clk_20_i)
  begin
    if rising_edge(clk_20_i) then
      if (rst_20_n = '0') then
        i2c_wdto_bit <= '0';
      elsif (i2c_wdto_p = '1') then
        i2c_wdto_bit <= '1';
      elsif (i2c_wdto_bit_rst_ld = '1') and (i2c_wdto_bit_rst = '1') then
        i2c_wdto_bit <= '0';
      end if;
    end if;
  end process p_sr_wdto_bit;

  -- Register for the I2C_ERR bit in the SR
  p_i2c_err_led : process(clk_20_i)
  begin
    if rising_edge(clk_20_i) then
      if (rst_20_n = '0') then
        i2c_err_bit <= '0';
      elsif (i2c_err_p = '1') then
        i2c_err_bit <= '1';
      elsif (i2c_err_bit_rst_ld = '1') and (i2c_err_bit_rst = '1') then
        i2c_err_bit <= '0';
      end if;
    end if;
  end process p_i2c_err_led;

  --============================================================================
  -- Instantiation and connection of the main Wishbone crossbar
  --============================================================================
  cmp_wb_crossbar : xwb_sdb_crossbar
    generic map
    (
      g_num_masters => c_nr_masters,
      g_num_slaves  => c_nr_slaves,
      g_registered  => false,
      g_wraparound  => true,
      g_layout      => c_sdb_layout,
      g_sdb_addr    => c_addr_sdb
    )
    port map
    (
      clk_sys_i => clk_20_i,
      rst_n_i   => rst_20_n,
      slave_i   => xbar_slave_in,
      slave_o   => xbar_slave_out,
      master_i  => xbar_master_in,
      master_o  => xbar_master_out
    );

  --============================================================================
  -- Converter board registers
  --============================================================================
  -- RTM lines combo
  rtm_lines    <= rtmp_i & rtmm_i;

--------------------------------------------------------------------------------
gen_line : if (g_nr_chans = c_max_nr_chans) generate
  line_front    <= line_front_i;
  line_rear     <= line_rear_i;
  line_front_fs <= line_front_fs_i;
  line_rear_fs  <= line_rear_fs_i;
end generate gen_line;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
gen_line_unused_chans : if (g_nr_chans < c_max_nr_chans) generate
  -- connect used lines
  line_front(g_nr_chans-1 downto 0)    <= line_front_i;
  line_rear(g_nr_chans-1 downto 0)     <= line_rear_i;
  line_front_fs(g_nr_chans-1 downto 0) <= line_front_fs_i;
  line_rear_fs(g_nr_chans-1 downto 0)  <= line_rear_fs_i;

  -- unused lines to zeroes
  line_front(c_max_nr_chans-1 downto g_nr_chans)    <= (others => '0');
  line_rear(c_max_nr_chans-1 downto g_nr_chans)     <= (others => '0');
  line_front_fs(c_max_nr_chans-1 downto g_nr_chans) <= (others => '0');
  line_rear_fs(c_max_nr_chans-1 downto g_nr_chans)  <= (others => '0');
end generate gen_line_unused_chans;
--------------------------------------------------------------------------------

  -- Implement the RST_UNLOCK bit
  p_rst_unlock : process (clk_20_i)
  begin
    if rising_edge(clk_20_i) then
      if (rst_20_n = '0') then
        rst_unlock <= '0';
      elsif (rst_unlock_bit_ld = '1') then
        if (rst_unlock_bit = '1') then
          rst_unlock <= '1';
        else
          rst_unlock <= '0';
        end if;
      end if;
    end if;
  end process p_rst_unlock;

  -- Implement the reset bit register
  -- The register can only be set when the RST_UNLOCK bit is '1'.
  p_rst_fr_reg : process (clk_20_i)
  begin
    if rising_edge(clk_20_i) then
      if (rst_20_n = '0') then
        rst_fr_reg <= '0';
      elsif (rst_bit_ld = '1') and (rst_bit = '1') and (rst_unlock = '1') then
        rst_fr_reg <= '1';
      else
        rst_fr_reg <= '0';
      end if;
    end if;
  end process p_rst_fr_reg;

  -- Register for the PMISSE bits in the SR, set when a channel misses a pulse
  -- Each bit is cleared by writing a '1' to it
  p_sr_pmisse_bit : process (clk_20_i)
  begin
    if rising_edge(clk_20_i) then
      for i in 0 to g_nr_chans-1 loop
        if (rst_20_n = '0') then
          pmisse_bit(i) <= '0';
        elsif (pmisse_p(i) = '1') then
          pmisse_bit(i) <= '1';
        elsif (pmisse_bit_rst_ld = '1') and (pmisse_bit_rst(i) = '1') then
          pmisse_bit(i) <= '0';
        end if;
      end loop;
    end if;
  end process p_sr_pmisse_bit;

  -- Create an OR of all PMISSE bits
  pmisse_bits_or <= '0' when (pmisse_bit = (pmisse_bit'range => '0')) else
                    '1';

--------------------------------------------------------------------------------
-- Set the rest of the PMISSE bits to zero when g_nr_chans < c_max_nr_chans
gen_pmisse_unused_chans : if (g_nr_chans < c_max_nr_chans) generate
  pmisse_bit(c_max_nr_chans-1 downto g_nr_chans) <= (others => '0');
end generate;
--------------------------------------------------------------------------------

  -- Synchronize WR valid signal to implement the WRPRES bit
  cmp_wrpres_sync : gc_sync_ffs
    generic map
    (
      g_sync_edge => "positive"
    )
    port map
    (
      clk_i    => clk_20_i,
      rst_n_i  => rst_20_n,
      data_i   => buf_wrtag,
      synced_o => wrpres
    );

  -- Implement the TBCSR.CLR bit
  p_tbcsr_clr : process (clk_20_i)
  begin
    if rising_edge(clk_20_i) then
      if (rst_20_n = '0') then
        buf_clr_p <= '0';
      else
        buf_clr_p <= '0';
        if (buf_clr_bit_ld = '1') and (buf_clr_bit = '1') then
          buf_clr_p <= '1';
        end if;
      end if;
    end if;
  end process p_tbcsr_clr;

  -- Implement the latest timestamp registers
  -- NOTE: Updated in the 125 MHz clock domain
  lts_ld_125 <= buf_chan;

  p_lts_125 : process (clk_125)
  begin
    if rising_edge(clk_125) then
      for i in 0 to g_nr_chans-1 loop
        if (rst_125_n = '0') then
          lts_cycles_125(i) <= (others => '0');
          lts_tai_125(i)    <= (others => '0');
          lts_wrtag_125(i)  <= '0';
        -- load only when synchronized version in 20 MHz domain is ready
        elsif (lts_ld_125(i) = '1') and (lts_ld_rdy_125(i) = '1') then
          lts_cycles_125(i) <= tm_cycles;
          lts_tai_125(i)    <= tm_tai;
          lts_wrtag_125(i)  <= buf_wrtag;
        end if;
      end loop;
    end if;
  end process p_lts_125;

  -- Pulse synchronizer: sync. lts_ld from 125 MHz to 20MHz domain
gen_lts_ld_pulse_sync : for i in 0 to g_nr_chans-1 generate
  cmp_pulse_sync : gc_pulse_synchronizer2
    port map
    (
      clk_in_i    => clk_125,
      rst_in_n_i  => rst_125_n,

      clk_out_i   => clk_20_i,
      rst_out_n_i => rst_20_n,

      d_ready_o   => lts_ld_rdy_125(i),

      d_p_i       => lts_ld_125(i),
      q_p_o       => lts_ld_20(i)
    );
end generate gen_lts_ld_pulse_sync;

  -- Latest timestamp regs in 20MHz clock domain
  p_lts_20 : process (clk_20_i)
  begin
    if rising_edge(clk_20_i) then
      for i in 0 to g_nr_chans-1 loop
        if (rst_20_n = '0') then
          lts_cycles_20(i) <= (others => '0');
          lts_tai_20(i)    <= (others => '0');
          lts_wrtag_20(i)  <= '0';
        elsif (lts_ld_20(i) = '1') then
          lts_cycles_20(i) <= lts_cycles_125(i);
          lts_tai_20(i)    <= lts_tai_125(i);
          lts_wrtag_20(i)  <= lts_wrtag_125(i);
        end if;
      end loop;
    end if;
  end process p_lts_20;

--------------------------------------------------------------------------------
  -- Connect unused timestamps to all zeroes
gen_latest_timestamp_unused_chans : if (g_nr_chans < c_max_nr_chans) generate

  lts_ld_125(c_max_nr_chans-1 downto g_nr_chans)     <= (others => '0');
  lts_ld_rdy_125(c_max_nr_chans-1 downto g_nr_chans) <= (others => '0');
  lts_cycles_125(c_max_nr_chans-1 downto g_nr_chans) <= (others => (others => '0'));
  lts_tai_125(c_max_nr_chans-1 downto g_nr_chans)    <= (others => (others => '0'));
  lts_wrtag_125(c_max_nr_chans-1 downto g_nr_chans)  <= (others => '0');

  lts_ld_20(c_max_nr_chans-1 downto g_nr_chans)      <= (others => '0');
  lts_cycles_20(c_max_nr_chans-1 downto g_nr_chans)  <= (others => (others => '0'));
  lts_tai_20(c_max_nr_chans-1 downto g_nr_chans)     <= (others => (others => '0'));
  lts_wrtag_20(c_max_nr_chans-1 downto g_nr_chans)   <= (others => '0');

end generate gen_latest_timestamp_unused_chans;
--------------------------------------------------------------------------------

  -- Then, instantiate the component
  cmp_conv_regs : conv_regs
    port map (
      rst_n_i                  => rst_20_n,
      clk_sys_i                => clk_20_i,

      wb_adr_i                 => xbar_master_out(c_slv_conv_regs).adr(7 downto 2),
      wb_dat_i                 => xbar_master_out(c_slv_conv_regs).dat,
      wb_dat_o                 => xbar_master_in (c_slv_conv_regs).dat,
      wb_cyc_i                 => xbar_master_out(c_slv_conv_regs).cyc,
      wb_sel_i                 => xbar_master_out(c_slv_conv_regs).sel,
      wb_stb_i                 => xbar_master_out(c_slv_conv_regs).stb,
      wb_we_i                  => xbar_master_out(c_slv_conv_regs).we,
      wb_ack_o                 => xbar_master_in (c_slv_conv_regs).ack,
      wb_stall_o               => xbar_master_in (c_slv_conv_regs).stall,

      reg_bidr_i               => g_board_id,

      reg_sr_gwvers_i          => g_gwvers,
      reg_sr_switches_i        => sw_gp_i,
      reg_sr_rtm_i             => rtm_lines,
      reg_sr_i2c_wdto_o        => i2c_wdto_bit_rst,
      reg_sr_i2c_wdto_i        => i2c_wdto_bit,
      reg_sr_i2c_wdto_load_o   => i2c_wdto_bit_rst_ld,
      reg_sr_wrpres_i          => wrpres,
      reg_sr_i2c_err_o         => i2c_err_bit_rst,
      reg_sr_i2c_err_i         => i2c_err_bit,
      reg_sr_i2c_err_load_o    => i2c_err_bit_rst_ld,
      reg_sr_pmisse_o          => pmisse_bit_rst,
      reg_sr_pmisse_i          => pmisse_bit,
      reg_sr_pmisse_load_o     => pmisse_bit_rst_ld,

      reg_cr_rst_unlock_o      => rst_unlock_bit,
      reg_cr_rst_unlock_i      => rst_unlock,
      reg_cr_rst_unlock_load_o => rst_unlock_bit_ld,
      reg_cr_rst_o             => rst_bit,
      reg_cr_rst_i             => rst_fr_reg,
      reg_cr_rst_load_o        => rst_bit_ld,
      reg_cr_mpt_o             => mpt,
      reg_cr_mpt_wr_o          => mpt_ld,

      reg_ch1pcr_o             => ch_pcr(0),
      reg_ch1pcr_i             => std_logic_vector(pulse_cnt(0)),
      reg_ch1pcr_load_o        => ch_pcr_ld(0),
      reg_ch2pcr_o             => ch_pcr(1),
      reg_ch2pcr_i             => std_logic_vector(pulse_cnt(1)),
      reg_ch2pcr_load_o        => ch_pcr_ld(1),
      reg_ch3pcr_o             => ch_pcr(2),
      reg_ch3pcr_i             => std_logic_vector(pulse_cnt(2)),
      reg_ch3pcr_load_o        => ch_pcr_ld(2),
      reg_ch4pcr_o             => ch_pcr(3),
      reg_ch4pcr_i             => std_logic_vector(pulse_cnt(3)),
      reg_ch4pcr_load_o        => ch_pcr_ld(3),
      reg_ch5pcr_o             => ch_pcr(4),
      reg_ch5pcr_i             => std_logic_vector(pulse_cnt(4)),
      reg_ch5pcr_load_o        => ch_pcr_ld(4),
      reg_ch6pcr_o             => ch_pcr(5),
      reg_ch6pcr_i             => std_logic_vector(pulse_cnt(5)),
      reg_ch6pcr_load_o        => ch_pcr_ld(5),

      reg_tvlr_o               => tvlr,
      reg_tvlr_i               => tm_tai(31 downto 0),
      reg_tvlr_load_o          => tvlr_ld,
      reg_tvhr_o               => tvhr,
      reg_tvhr_i               => tm_tai(39 downto 32),
      reg_tvhr_load_o          => tvhr_ld,

      reg_tbmr_chan_i          => buf_dat_out( 5 downto  0),
      reg_tbmr_wrtag_i         => buf_dat_out( 6),
      reg_tb_rd_req_p_o        => buf_rd_req_p,
      reg_tbcyr_i              => buf_dat_out(34 downto  7),
      reg_tbtlr_i              => buf_dat_out(66 downto 35),
      reg_tbthr_i              => buf_dat_out(74 downto 67),
      reg_tbcsr_clr_o          => buf_clr_bit,
      reg_tbcsr_clr_i          => '0',
      reg_tbcsr_clr_load_o     => buf_clr_bit_ld,
      reg_tbcsr_usedw_i        => buf_count,
      reg_tbcsr_full_i         => buf_full,
      reg_tbcsr_empty_i        => buf_empty,

      reg_ch1ltscyr_i          => lts_cycles_20(0),
      reg_ch1ltstlr_i          => lts_tai_20(0)(31 downto  0),
      reg_ch1ltsthr_tai_i      => lts_tai_20(0)(39 downto 32),
      reg_ch1ltsthr_wrtag_i    => lts_wrtag_20(0),
      reg_ch2ltscyr_i          => lts_cycles_20(1),
      reg_ch2ltstlr_i          => lts_tai_20(1)(31 downto  0),
      reg_ch2ltsthr_tai_i      => lts_tai_20(1)(39 downto 32),
      reg_ch2ltsthr_wrtag_i    => lts_wrtag_20(1),
      reg_ch3ltscyr_i          => lts_cycles_20(2),
      reg_ch3ltstlr_i          => lts_tai_20(2)(31 downto  0),
      reg_ch3ltsthr_tai_i      => lts_tai_20(2)(39 downto 32),
      reg_ch3ltsthr_wrtag_i    => lts_wrtag_20(2),
      reg_ch4ltscyr_i          => lts_cycles_20(3),
      reg_ch4ltstlr_i          => lts_tai_20(3)(31 downto  0),
      reg_ch4ltsthr_tai_i      => lts_tai_20(3)(39 downto 32),
      reg_ch4ltsthr_wrtag_i    => lts_wrtag_20(3),
      reg_ch5ltscyr_i          => lts_cycles_20(4),
      reg_ch5ltstlr_i          => lts_tai_20(4)(31 downto  0),
      reg_ch5ltsthr_tai_i      => lts_tai_20(4)(39 downto 32),
      reg_ch5ltsthr_wrtag_i    => lts_wrtag_20(4),
      reg_ch6ltscyr_i          => lts_cycles_20(5),
      reg_ch6ltstlr_i          => lts_tai_20(5)(31 downto  0),
      reg_ch6ltsthr_tai_i      => lts_tai_20(5)(39 downto 32),
      reg_ch6ltsthr_wrtag_i    => lts_wrtag_20(5),

      reg_lsr_front_i          => line_front,
      reg_lsr_frontinv_i       => line_inv_i,
      reg_lsr_rear_i           => line_rear,
      reg_lsr_frontfs_i        => line_front_fs,
      reg_lsr_frontinvfs_i     => line_inv_fs_i,
      reg_lsr_rearfs_i         => line_rear_fs,

      reg_oswr_switches_i      => sw_other_i,

      reg_tempr_i              => tempr,
      reg_uidlr_i              => uidr(31 downto  0),
      reg_uidhr_i              => uidr(63 downto 32)
    );

  --============================================================================
  -- Instantiate Xilinx MultiBoot module
  --============================================================================
  cmp_multiboot : xwb_xil_multiboot
    port map
    (
      clk_i      => clk_20_i,
      rst_n_i    => rst_20_n,
      wbs_i      => xbar_master_out(c_slv_multiboot),
      wbs_o      => xbar_master_in(c_slv_multiboot),
      spi_cs_n_o => flash_cs_n_o,
      spi_sclk_o => flash_sclk_o,
      spi_mosi_o => flash_mosi_o,
      spi_miso_i => flash_miso_i
    );

  --============================================================================
  -- On-board DS18B20 Thermometer logic
  --============================================================================
--------------------------------------------------------------------------------
gen_thermometer : if (g_with_thermometer = true) generate
  cmp_therm : serialIdTempInt
    generic map (
      FREQ => 20
      )
    port map (
      Clk      => clk_20_i,
      RstN     => rst_20_n,

      SerialId => thermometer_b,
      Id       => uidr,
      Temp     => tempr,
      IdRead   => open,
      Pps      => pps_p,
      IdOk     => open
      );
end generate gen_thermometer;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
gen_no_thermometer : if (g_with_thermometer = false) generate
  thermometer_b <= 'Z';
end generate gen_no_thermometer;
--------------------------------------------------------------------------------

  --============================================================================
  -- Bicolor LED matrix logic
  --============================================================================
  -- connect column, line & state signals outside
  cmp_bicolor_led_ctrl : gc_bicolor_led_ctrl
    generic map
    (
      g_NB_COLUMN    => g_bicolor_led_columns,
      g_NB_LINE      => g_bicolor_led_lines,
      g_clk_freq     => 20000000,
      g_refresh_rate => 250
    )
    port map
    (
      clk_i           => clk_20_i,
      rst_n_i         => rst_20_n,
      led_intensity_i => "1111111",
      led_state_i     => bicolor_led_state_i,
      column_o        => bicolor_led_col_o,
      line_o          => bicolor_led_line_o,
      line_oen_o      => bicolor_led_line_oen_o
    );

  -- Set the system error signal for the ERR LED
  led_syserr_o <= i2c_err_bit or i2c_wdto_bit or pmisse_bits_or;

  --============================================================================
  -- White Rabbit
  --============================================================================
--------------------------------------------------------------------------------
gen_with_wr : if (g_with_wr = true) generate

  -- WR-specific clocking
  cmp_clk_20_buf : BUFG
    port map
    (
      O => clk_20_buf,
      I => clk_20_i
    );

  cmp_dmtd_clk_pll : PLL_BASE
    generic map (
      BANDWIDTH          => "OPTIMIZED",
      CLK_FEEDBACK       => "CLKFBOUT",
      COMPENSATION       => "INTERNAL",
      DIVCLK_DIVIDE      => 1,
      CLKFBOUT_MULT      => 50,
      CLKFBOUT_PHASE     => 0.000,
      CLKOUT0_DIVIDE     => 16,         -- 62.5 MHz
      CLKOUT0_PHASE      => 0.000,
      CLKOUT0_DUTY_CYCLE => 0.500,
      CLKIN_PERIOD       => 50.0,
      REF_JITTER         => 0.016
    )
    port map (
      CLKFBOUT => clk_dmtd_pll_fb,
      CLKOUT0  => clk_dmtd_fr_pll,
      RST      => '0',
      CLKFBIN  => clk_dmtd_pll_fb,
      CLKIN    => clk_20_i
    );

  cmp_clk_dmtd_buf : BUFG
    port map (
      O => clk_dmtd,
      I => clk_dmtd_fr_pll
    );

  -- Instantiate the White Rabbit PTP Core (WRPC)
  c_wr_core : xwr_core
    generic map (
      g_simulation                => 0,
      g_with_external_clock_input => false,
      --
      g_phys_uart                 => true,
      g_virtual_uart              => false,
      g_aux_clks                  => 1,
      g_ep_rxbuf_size             => 1024,
      g_dpram_initf               => "", --"wrc.ram",
      g_dpram_size                => 90112/4,  --16384,
      g_interface_mode            => PIPELINED,
      g_address_granularity       => BYTE
    )
    port map (
      clk_sys_i  => clk_20_i,
      clk_dmtd_i => clk_dmtd,
      clk_ref_i  => clk_125,
      clk_aux_i  => (others => '0'),
      clk_ext_i  => '0',
      pps_ext_i  => '0',
      rst_n_i    => rst_20_n,

      dac_hpll_load_p1_o => dac_hpll_load_p,
      dac_hpll_data_o    => dac_hpll_data,
      dac_dpll_load_p1_o => dac_dpll_load_p,
      dac_dpll_data_o    => dac_dpll_data,

      phy_ref_clk_i      => clk_125,
      phy_tx_data_o      => phy_tx_data,
      phy_tx_k_o         => phy_tx_k,
      phy_tx_disparity_i => phy_tx_disparity,
      phy_tx_enc_err_i   => phy_tx_enc_err,
      phy_rx_data_i      => phy_rx_data,
      phy_rx_rbclk_i     => phy_rx_rbclk,
      phy_rx_k_i         => phy_rx_k,
      phy_rx_enc_err_i   => phy_rx_enc_err,
      phy_rx_bitslide_i  => phy_rx_bitslide,
      phy_rst_o          => phy_rst,
      phy_loopen_o       => phy_loopen,

      led_act_o   => wr_led_act_o,
      led_link_o  => wr_led_link_o,
      scl_o       => open,
      scl_i       => '1',
      sda_o       => open,
      sda_i       => '1',
      sfp_scl_o   => sfp_scl_fr_wrpc,
      sfp_scl_i   => sfp_scl_to_wrpc,
      sfp_sda_o   => sfp_sda_fr_wrpc,
      sfp_sda_i   => sfp_sda_to_wrpc,
      sfp_det_i   => sfp_present_i,
      btn1_i      => '0',
      btn2_i      => '0',

      uart_rxd_i => '1',
      uart_txd_o => open,

      owr_en_o => open,
      owr_i    => (others => '1'),

      slave_i => cc_dummy_slave_in,
      slave_o => open,

      aux_master_o => open,
      aux_master_i => cc_dummy_master_in,

--      wrf_src_o => open,
--      wrf_src_i => c_dummy_src_in,
--      wrf_snk_o => open,
--      wrf_snk_i => c_dummy_snk_in,

      tm_dac_value_o       => open,
      tm_dac_wr_o          => open,
      tm_clk_aux_lock_en_i => (others=>'0'),
      tm_clk_aux_locked_o  => open,
      tm_time_valid_o      => open,
      tm_tai_o             => open,
      tm_cycles_o          => open,
      pps_p_o              => pps_p,
      pps_led_o            => open,

      dio_o       => open,
      rst_aux_n_o => open
    );

  -- Connect the SFP's I2C interface to the WRPC signals
  sfp_scl_b       <= '0' when sfp_scl_fr_wrpc = '0' else 'Z';
  sfp_sda_b       <= '0' when sfp_sda_fr_wrpc = '0' else 'Z';
  sfp_scl_to_wrpc <= sfp_scl_b;
  sfp_sda_to_wrpc <= sfp_sda_b;

  sfp_tx_disable_o  <= '0';
  sfp_rate_select_o <= '0';

  -- Instantiate the PHY MGT and its dedicated clock buffer
  cmp_clk_gtp : IBUFGDS
    generic map
    (
      DIFF_TERM    => true,
      IBUF_LOW_PWR => true,
      IOSTANDARD   => "DEFAULT"
    )
    port map
    (
      O  => clk_gtp,
      I  => mgt_refclk0_p_i,
      IB => mgt_refclk0_n_i
    );

  cmp_gtp : wr_gtp_phy_spartan6
    generic map
    (
      g_simulation      => 0,
      g_enable_ch0      => 1,
      g_enable_ch1      => 0
    )
    port map
    (
      gtp_clk_i          => clk_gtp,
      ch0_ref_clk_i      => clk_125,
      ch0_tx_data_i      => phy_tx_data(7 downto 0),
      ch0_tx_k_i         => phy_tx_k,
      ch0_tx_disparity_o => phy_tx_disparity,
      ch0_tx_enc_err_o   => phy_tx_enc_err,
      ch0_rx_rbclk_o     => phy_rx_rbclk,
      ch0_rx_data_o      => phy_rx_data(7 downto 0),
      ch0_rx_k_o         => phy_rx_k,
      ch0_rx_enc_err_o   => phy_rx_enc_err,
      ch0_rx_bitslide_o  => phy_rx_bitslide(3 downto 0),
      ch0_rst_i          => phy_rst,
      ch0_loopen_i       => phy_loopen,

      ch1_ref_clk_i      => clk_125,
      ch1_tx_data_i      => x"00",
      ch1_tx_k_i         => '0',
      ch1_tx_disparity_o => open,
      ch1_tx_enc_err_o   => open,
      ch1_rx_rbclk_o     => open,
      ch1_rx_data_o      => open,
      ch1_rx_k_o         => open,
      ch1_rx_enc_err_o   => open,
      ch1_rx_bitslide_o  => open,
      ch1_rst_i          => '1',
      ch1_loopen_i       => '0',

      pad_txn0_o         => mgt_tx0_n_o,
      pad_txp0_o         => mgt_tx0_p_o,
      pad_rxn0_i         => mgt_rx0_n_i,
      pad_rxp0_i         => mgt_rx0_p_i,

      pad_txn1_o         => open,
      pad_txp1_o         => open,
      pad_rxn1_i         => '0',
      pad_rxp1_i         => '0'
    );

  -- DAC drivers
  cmp_dac20_driver : spec_serial_dac_arb
    generic map
    (
      g_invert_sclk    => false,
      g_num_extra_bits => 8
    )
    port map
    (
      clk_i   => clk_20_i,
      rst_n_i => rst_20_n,

      val1_i  => dac_hpll_data,
      load1_i => dac_hpll_load_p,

      val2_i  => (others => '0'),
      load2_i => '0',

      dac_cs_n_o(0) => dac20_sync_n_o,
      dac_cs_n_o(1) => open,
      dac_sclk_o    => dac20_sclk_o,
      dac_din_o     => dac20_din_o
    );

  cmp_dac125_driver : spec_serial_dac_arb
    generic map
    (
      g_invert_sclk    => false,
      g_num_extra_bits => 8
    )
    port map
    (
      clk_i   => clk_20_i,
      rst_n_i => rst_20_n,

      val1_i  => dac_dpll_data,
      load1_i => dac_dpll_load_p,

      val2_i  => (others => '0'),
      load2_i => '0',

      dac_cs_n_o(0) => dac125_sync_n_o,
      dac_cs_n_o(1) => open,
      dac_sclk_o    => dac125_sclk_o,
      dac_din_o     => dac125_din_o
    );

end generate gen_with_wr;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
gen_no_wr : if (g_with_wr = false) generate

  -- DAC outputs: enables to '1' (disable DAC comm interface) and SCK, DIN to '0'
  dac20_sync_n_o  <= '1';
  dac20_din_o     <= '0';
  dac20_sclk_o    <= '0';
  dac125_sync_n_o <= '1';
  dac125_din_o    <= '0';
  dac125_sclk_o   <= '0';

  -- SFP lines all open-drain, set to high-impedance
  sfp_rate_select_o <= 'Z';
  sfp_scl_b         <= 'Z';
  sfp_sda_b         <= 'Z';
  sfp_tx_disable_o  <= 'Z';

  -- Implement PPS counter for the thermometer module (normally from WR)
  gen_therm_pps_no_wr : if (g_with_thermometer = true) generate
    p_therm_pps : process (clk_20_i)
    begin
      if rising_edge(clk_20_i) then
        if (rst_20_n = '0') then
          pps_cnt <= (others => '0');
          pps_p   <= '0';
        else
          pps_p   <= '0';
          pps_cnt <= pps_cnt + 1;
          if (pps_cnt = 19999999) then
            pps_p   <= '1';
            pps_cnt <= (others => '0');
          end if;
        end if;
      end if;
    end process p_therm_pps;
  end generate gen_therm_pps_no_wr;

end generate gen_no_wr;
--------------------------------------------------------------------------------

end architecture arch;
--==============================================================================
--  architecture end
--==============================================================================
